#ifdef _CH_
#pragma package <opencv>
#endif

#include "ipp.h"
#include "cv.h"
#include "highgui.h"
#include <time.h>
#include <math.h>
#include <ctype.h>
#include <stdio.h>

IplImage* image = 0;
IplImage* image_overlay = 0;
IplImage* image_gt= 0;
IplImage* image_alg= 0;
IplImage* image_alg_grid= 0;

IplImage *image_tmp_float;
bool next_frame=true;

int is_interactive=1;

#define GET_PIXEL(IMG,TYPE,X,Y) ((TYPE)((IMG)->imageData + (IMG)->widthStep*(Y)))[X]
#define GET_PIXEL_COL(IMG,TYPE,NUM_COLS,X,Y,C) ((TYPE)((IMG)->imageData + (IMG)->widthStep*(Y)))[NUM_COLS*X+C]



/******************************************************************************************/
bool neighIsTP(IplImage *img, int x, int y, int threshold) {
	int coord_x[] = {-1,-1,-1, 0, 0, 0, 1, 1, 1};
	int coord_y[] = {-1, 0, 1,-1, 0, 1,-1, 0, 1};
	for(int i=0;i<9;i++) {
		int xx=(x+coord_x[i]+img->width)%img->width;
		int yy=(y+coord_y[i]+img->height)%img->height;
		if (GET_PIXEL(img,unsigned char*,xx,yy)>=threshold)
			return true;
	}
	return false;
}


/******************************************************************************************/
void decorateImage(IplImage *img_in, IplImage *img_gt, IplImage *img_res, IplImage *img_out, int cut_off) {
	if (img_in) cvResize(img_in,img_out);
	else cvSet(img_out,cvScalarAll(0));
	
	float step_x= (img_out->width*1.0f/img_gt->width);
	float step_y= (img_out->height*1.0f/img_gt->height);

	for(int y=0;y<img_out->height;y++)
		for(int x=0;x<img_out->width;x++) {
			int grid_x=(int)floor(x/step_x);
			int grid_y=(int)floor(y/step_y);

			int val_gt= GET_PIXEL(img_gt,unsigned char*,grid_x,grid_y);
			int val_alg= GET_PIXEL(img_res,unsigned char*,grid_x,grid_y);
			val_alg=(val_alg>=cut_off)?255:0;

			if (val_gt==val_alg) {
				if (val_gt!=0)
					GET_PIXEL_COL(img_out,unsigned char*,3,x,y,1)=255; // match - green

			}
			else {
				if (val_gt!=0) {
//					if (neighIsTP(img_res,x/step_x,y/step_y)==false) 
						GET_PIXEL_COL(img_out,unsigned char*,3,x,y,0)=255; //false negative - blue
//					else
//						GET_PIXEL_COL(img_out,unsigned char*,3,x,y,1)=255; // neigh match - green
				}
				else {
//					if (neighIsTP(img_gt,x/step_x,y/step_y)==false) 
						GET_PIXEL_COL(img_out,unsigned char*,3,x,y,2)=255; //false positive - red
//					else
//						GET_PIXEL_COL(img_out,unsigned char*,3,x,y,1)=255; // neigh match - green


				}
			}
				
		}

	for(int y=0;y<img_gt->height;y++) {
			CvPoint pt1,pt2;
			pt1.x = 1;
			pt1.y = (int)floor(y*step_y);
			pt2.x = img_out->width;
			pt2.y = (int)floor(y*step_y);
			cvLine( img_out, pt1, pt2 , CV_RGB(255,0,0), 1, CV_AA, 0 );
	}

	for(int x=0;x<img_gt->width;x++) {
			CvPoint pt1,pt2;
			pt1.x = (int)floor(x*step_x);
			pt1.y = 1;
			pt2.x = (int)floor(x*step_x);
			pt2.y = img_out->height;
			cvLine( img_out, pt1, pt2 , CV_RGB(255,0,0), 1, CV_AA, 0 );
	}

}


/******************************************************************************************/
void markGroundtruth( int event, int x, int y, int flags, void* param )
{
    if( !image )return;

    if( image->origin )
        y = image->height - y;

    if( event == CV_EVENT_LBUTTONDOWN )
    {
		int gt_x= x / (image->width/image_gt->width);
		int gt_y= y / (image->height/image_gt->height);
		
		unsigned char current_value = GET_PIXEL(image_gt,unsigned char*,gt_x,gt_y);
		GET_PIXEL(image_gt,unsigned char*,gt_x,gt_y) = 255 - current_value;
    }
	else
	if( event == CV_EVENT_RBUTTONDOWN )
    {
		next_frame=true;
	}

}

/******************************************************************************************/
void evaluateData(IplImage *img_gt, IplImage *img_res, int *true_pos, int *false_pos, int *false_neg, int threshold) {
	int tp=0;
	int fp=0;
	int fn=0;
	for(int y=0;y<img_gt->height;y++)
		for(int x=0;x<img_gt->width;x++) {
			int val_gt= GET_PIXEL(img_gt,unsigned char*,x,y);
			int val_alg= GET_PIXEL(img_res,unsigned char*,x,y);
			
			val_alg=(val_alg>threshold)?255:0;

			if (val_gt==val_alg) {
				if (val_gt!=0)
					tp++;
			}
			else {
				if (val_gt!=0) {
					if (neighIsTP(img_res,x,y,threshold)==false) 
						fn++;
					else
						tp++;
				}
				else {
					if (neighIsTP(img_gt,x,y,255)==false) 
						fp++;
					else
						tp++;
				}
					
			}
		}
		*true_pos = *true_pos + tp;
		*false_pos = *false_pos + fp;
		*false_neg = *false_neg + fn;
}

/******************************************************************************************/
void generateGridAlgOutput( IplImage * img_alg, IplImage *img_grid) 
{
	cvSet(image_tmp_float,cvScalarAll(0));

	int step_x= (int)ceil(img_alg->width*1.0f/img_grid->width);
	int step_y= (int)ceil(img_alg->height*1.0f/img_grid->height);
	float normalization = (float)step_x*step_y;

	for(int y=0;y<img_alg->height;y++)
		for(int x=0;x<img_alg->width;x++) {
			GET_PIXEL(image_tmp_float,float*,x/step_x,y/step_y) =  GET_PIXEL(img_alg,unsigned char*,x,y)/normalization + GET_PIXEL(image_tmp_float,float*,x/step_x,y/step_y);
		}

	for(int y=0;y<img_grid->height;y++)
		for(int x=0;x<img_grid->width;x++) {
			unsigned char tmp_value = 0;
			//if ( GET_PIXEL(image_tmp_float,float*,x,y) >= threshold) tmp_value = 255  ;	
			tmp_value = MIN(255,(unsigned char) ceil(GET_PIXEL(image_tmp_float,float*,x,y)));
			GET_PIXEL(img_grid,unsigned char*,x,y) =  tmp_value;
		}
}

/******************************************************************************************/
int main(int argc, char** argv) {            
		
	char loadFilename[1010];

    CvCapture* capture = 0;
	int frame_counter=0;

//	char save_dir[1000]="";
	
	char *input_gt_filemask=NULL;
	char *input_alg_filemask=NULL;
	char *output_filemask=NULL;
	int input_shift=0;

	int frame_start=0;
	int frame_end=99999999;
	int frame_skip=1;

	int	cut_off=100;
	char *input_file=NULL;

	bool is_init=false;	
	bool step_by_step=false;	

	//interactive values
	int tp=0,fp=0,fn=0;

	const int no_values=50;
	int v_tp[no_values];
	int v_fp[no_values];
	int v_fn[no_values];

	if (( argc<= 2) || (strcmp(argv[1],"/?")==0)) {
		printf("Usage groundtruth [source_gt] [source_alg] [grid_cols]\n");
		printf("\t source_gt    = filemask for groundtruth (i.e. gt-%04d.png)\n");
		printf("\t source_alg   = filemask for alg. output (i.e. out-%04d.png)\n");
		printf("\t source_shift = shift between numbering source schemes \n");
		printf("\t                source_gt(i) = source_alg(i+source_shift)\n");
		printf("\t frame_start  = frame start\n");
		printf("\t frame_end    = frame end\n");
		printf("\t frame_skip   = frame skip\n");
		printf("---- if additional params are given it will start in interactive mode -----\n");
		printf("\t cut_off      = threshold for display purposes (btw 0-255)\n");
		printf("\t source_file  = input video file\n");
		printf("---- if additional params are given output will be saved-----\n");
		printf("\t dest_output  = filemask on where to save the output (i.e. eval-out-%04d.png)\n");
		printf("\t modes	    = 0 - no interactive (all outputs generated); 1 - full-interactive; -1 - only PR data saved");
		
		printf("--------------------------------------------------------------\n");
		printf("When in interactive mode, the following keys are enabled:\n");
		printf("\t - 's' - saves a snapshot of the current image\n");
		printf("\t - ' ' - toggles pauses \n");
		printf("\t - 'n' - toggles step-by-step viewing\n");
		printf("\t - ESC - exits the program\n");

		return 0;
	}

	int p=1;

	p++;
	if (argc >= p) {
		input_gt_filemask = argv[p-1];
		printf("Groundtruth filemask: %s\n",input_gt_filemask );
	}

	p++;
	if (argc >= p) {
		input_alg_filemask = argv[p-1];
		printf("Algorithm filemask: %s\n",input_alg_filemask );
	}

	p++;
	if (argc >= p) {
		input_shift=atoi(argv[p-1]);
		printf("Source Shift: %d\n",input_shift);
	}

	p++;
	if (argc >= p) {
		frame_start=atoi(argv[p-1]);
		printf("Frame start: %d\n",frame_start);
	}

	p++;
	if (argc >= p) {
		frame_end=atoi(argv[p-1]);
		printf("Frame end: %d\n",frame_end);
	}

	p++;
	if (argc >= p) {
		frame_skip=atoi(argv[p-1]);
		printf("Frame skip: %d\n",frame_skip);
	}

	p++;
	if (argc >= p) {
		cut_off=atoi(argv[p-1]);
		printf("Cut off: %d\n",cut_off);
	}

	p++;
	if (argc >= p) {
		input_file = argv[p-1];
		printf("Source file: %s\n",input_file);
	}

	p++;
	if (argc >= p) {
		output_filemask = argv[p-1];
		printf("Output filemask: %s\n",output_filemask );
	}

	p++;
	if (argc >= p) {
		is_interactive=atoi(argv[p-1]);
		printf("interactive: %d\n",is_interactive);
	}


	//initialize the groundtruth
	for(int v=0;v<no_values;v++) {
		v_tp[v]=0;
		v_fp[v]=0;
		v_fn[v]=0;
	}

	if (is_interactive>=0) {
		capture = cvCaptureFromFile( input_file );
		if (is_interactive>0) cvNamedWindow( "Image", 1 );
	}
	//cvSetMouseCallback( "Image", markGroundtruth, 0 );

	frame_counter = 0;
    for(;frame_counter<frame_end;) {

		if (next_frame) {
			frame_counter++;

			if( capture ) {
				bool error=false;
				for(int skip=0;skip<frame_skip;skip++)	{
					if( !cvGrabFrame( capture )) {
						error=true;
						break;
					}
					image = cvRetrieveFrame( capture );
				}

				if (error) {
					printf("Error grabbing frame\n");
					break;
				}
			}

			if (frame_counter < frame_start) {
				printf("\rFrame=%05d ",frame_counter);
				continue;
			}

			// load groundtruth
			if (input_gt_filemask) {
				sprintf(loadFilename,input_gt_filemask,frame_counter);
				IplImage *img_gt_new = cvLoadImage(loadFilename,0);
				if (img_gt_new) {
					if (image_gt) cvReleaseImage(&image_gt);
					image_gt = img_gt_new;
				}
				else {
					printf("Could not load groundtruth file %s\n",loadFilename);
					break;
				}
			}

			// load algorithm output
			if (input_alg_filemask) {
				sprintf(loadFilename,input_alg_filemask,frame_counter+input_shift);
				IplImage *img_alg_new = cvLoadImage(loadFilename,0);
				if (img_alg_new) {
					if (image_alg) cvReleaseImage(&image_alg);
					image_alg= img_alg_new;
				}
				else {
					printf("Could not load algorithm output file %s",loadFilename);
					break;
				}
			}
		}

		if (is_init==false) {
			if (capture)
				image_overlay = cvCreateImage(cvSize(image->width,image->height), IPL_DEPTH_8U, 3);
			else {
				if (!image_alg) {
					printf("First image_alg could not be loaded!");
					exit(-1);
				}
				image_overlay = cvCreateImage(cvSize(image_alg->width,image_alg->height), IPL_DEPTH_8U, 3);
				image = cvCreateImage(cvSize(image_alg->width,image_alg->height), IPL_DEPTH_8U, 3);
			}

			image_alg_grid = (IplImage*)cvClone(image_gt);
			image_tmp_float = cvCreateImage(cvSize(image_gt->width,image_gt->height), IPL_DEPTH_32F,1);
			is_init=true;
		}

		if (step_by_step) next_frame=false;

		if (next_frame) {
			generateGridAlgOutput(image_alg,image_alg_grid);

			if (is_interactive>=0) {
				evaluateData(image_gt,image_alg_grid,&tp,&fp,&fn,cut_off);
				decorateImage(image,image_gt,image_alg_grid,image_overlay,cut_off);
			}
		}

		
		if(is_interactive>0) {

			cvShowImage( "Image", image_overlay);

			int keyPressed = cvWaitKey(1);
			if (keyPressed  == 27) break;
			if (keyPressed  == ' ') {
				next_frame=!next_frame;
				step_by_step=false;
			}
			if (keyPressed == 'n' || keyPressed == 'N') {
				step_by_step = true;
				next_frame=true;
			}
			
			if (keyPressed == 's' || keyPressed == 'S') {
				char saveFilename[1010];
				sprintf(saveFilename,"snapshot-frame-%05d.png",frame_counter);
				printf("\rSaving local snapshot to file %s\n",saveFilename);
				cvSaveImage(saveFilename,image_overlay);

			}
		} 

		if (output_filemask && (is_interactive>=0)) {

				char saveFilename[1010];
				sprintf(saveFilename,output_filemask,frame_counter);
				//printf("\rSaving %s",saveFilename);
				cvSaveImage(saveFilename,image_overlay);
		}
		printf("\rFrame=%05d tp=%d fp=%d fn=%d",frame_counter,tp,fp,fn);

		if (next_frame) {

			for(int v=0;v<no_values;v++) {
				int tmp_cut_off=255-255*v/(no_values-1);
				evaluateData(image_gt,image_alg_grid,&v_tp[v],&v_fp[v],&v_fn[v],tmp_cut_off);

			}
		}

	}

	printf("\nROC graph: data threshold => (fp_rate,tp_rate):\n");
	for(int v=0;v<no_values;v++) {
		int tmp_cut_off=255-255*v/(no_values-1);
		int gt_tp= v_tp[v]+v_fn[v]+1;
		int gt_fp= (image_gt->width*image_gt->height)*(frame_counter-frame_start) - v_tp[v] - v_fn[v];

		float tp_rate=v_tp[v]*1.0f/gt_tp;
		float fp_rate=v_fp[v]*1.0f/gt_fp;

		printf("%03d => (%.3f %.3f)\n",tmp_cut_off,fp_rate,tp_rate);
	}


	printf("\nRecall-Precision graph: data threshold => (recall,precision):\n");
	FILE *the_file;

	char recallFilename[1010];
	if (output_filemask) {
		sprintf(recallFilename,output_filemask,0);
		sprintf(recallFilename,"%s.recall-precision.txt",recallFilename);
	}
	else
		sprintf(recallFilename,"%recall-precision.txt");

	the_file=fopen(recallFilename,"wt");
	for(int v=0;v<no_values;v++) {
		int tmp_cut_off=255-255*v/(no_values-1);
		int gt_tp= v_tp[v]+v_fn[v]+1;

		float precision_rate=(v_tp[v]+0.0001f)/(v_tp[v] + v_fp[v]+0.0001f);
		float recall_rate=MIN(1.0f,(v_tp[v]+0.0001f)/(gt_tp+0.0001f));

		printf("%03d => (%.3f %.3f)\n",tmp_cut_off,recall_rate,precision_rate);
		fprintf(the_file,"%f %f\n",recall_rate,precision_rate);
	}
	fclose(the_file);

	// gotto release the basis and the buffer
	if (is_interactive>=0) {
		cvReleaseCapture( &capture );
		if (is_interactive>0) cvDestroyWindow( "Image");
	}


	cvReleaseImage(&image_overlay);
	if (image_gt) cvReleaseImage(&image_gt);
	if (image_alg) cvReleaseImage(&image_alg);
	if (image_overlay) cvReleaseImage(&image_overlay);
	if (image_alg_grid) cvReleaseImage(&image_alg_grid);

	printf("\nExiting\n");
    return 0;
}